# -*- coding: utf-8 -*- #
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configure build and deploy to Google Kubernetes Engine command."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import re

from apitools.base.py.exceptions import HttpNotFoundError
from googlecloudsdk.api_lib.cloudbuild import cloudbuild_util
from googlecloudsdk.api_lib.source import sourcerepo
from googlecloudsdk.api_lib.storage import storage_api
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.calliope import actions
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as c_exceptions
from googlecloudsdk.command_lib.builds import staging_bucket_util
from googlecloudsdk.command_lib.builds.deploy import build_util
from googlecloudsdk.core import exceptions as core_exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core import resources

import six

_CLEAN_PREVIEW_SCHEDULE_HOUR = 12
_CLEAN_PREVIEW_SCHEDULE = '0 {} * * *'.format(_CLEAN_PREVIEW_SCHEDULE_HOUR)
_REPO_TYPE_CODES = {
    'github': 'ga',
    'bitbucket_mirrored': 'bm',
    'github_mirrored': 'gm',
    'csr': 'cs'
}


class SourceRepoNotConnectedException(c_exceptions.InvalidArgumentException):
  """Exception for when a third party CSR repo is not found.

  This should not be used for regular CSR repos since they are connected by
  default.
  """

  def __init__(self, parameter_name, message):
    super(SourceRepoNotConnectedException, self).__init__(
        parameter_name,
        '{message}\n\n'
        'Visit https://console.cloud.google.com/cloud-build/triggers/connect?project={project} '
        'to connect a repository to your project.'.format(
            message=message,
            project=properties.VALUES.core.project.Get(required=True),
        ))


class ConfigureGKEDeploy(base.Command):
  """Configure automated build and deployment to a target Google Kubernetes Engine cluster.

  Configure automated build and deployment from a repository. This can be
  triggered by a Git branch or tag push.
  """

  @staticmethod
  def Args(parser):
    """Register flags for this command.

    Args:
      parser: An argparse.ArgumentParser-like object. It is mocked out in order
        to capture some information, but behaves like an ArgumentParser.
    """
    parser.add_argument(
        '--repo-type',
        help="""
        Type of repository.

        `--repo-owner` must be provided if one of the following choices is
        selected:

        `github` - A GitHub (Cloud Build GitHub App) repository connected to
        Cloud Build triggers. The deployed image will have the format
        'gcr.io/[PROJECT_ID]/github.com/[REPO_OWNER]/[REPO_NAME]:$COMMIT_SHA'.

        `bitbucket_mirrored` - A Bitbucket repository connected to Cloud Source
        Repositories. The deployed image will have the format
        'gcr.io/[PROJECT_ID]/bitbucket.org/[REPO_OWNER]/[REPO_NAME]:$COMMIT_SHA'.

        `github_mirrored` - A GitHub repository connected to Cloud Source
        Repositories. The deployed image will have the format
        'gcr.io/[PROJECT_ID]/github.com/[REPO_OWNER]/[REPO_NAME]:$COMMIT_SHA'.

        `--repo-owner` must not be provided if the following is selected:

        `csr` - A repository on Cloud Source Repositories. The deployed image
        will have the format 'gcr.io/[PROJECT_ID]/[REPO_NAME]:$COMMIT_SHA'.

        Connect repositories at
        https://console.cloud.google.com/cloud-build/triggers/connect.
        """,
        choices=['github', 'bitbucket_mirrored', 'github_mirrored', 'csr'],
        required=True)
    parser.add_argument(
        '--repo-name',
        help='Name of the repository.',
        required=True)
    parser.add_argument(
        '--repo-owner',
        help='Owner of the repository.')
    parser.add_argument(
        '--dockerfile',
        help="""
        Path to the Dockerfile to build from, relative to the repository.

        Defaults to './Dockerfile'.
        """)
    trigger_match = parser.add_mutually_exclusive_group(required=True)
    trigger_match.add_argument(
        '--branch-pattern',
        metavar='REGEX',
        help='''
        A regular expression specifying which Git branches to match.

        This pattern is used as a regex search for any incoming pushes. For
        example, --branch-pattern=foo will match "foo", "foobar", and "barfoo".
        Events on a branch that does not match will be ignored.

        The syntax of the regular expressions accepted is the syntax accepted by
        RE2 and described at https://github.com/google/re2/wiki/Syntax.
        ''')
    trigger_match.add_argument(
        '--tag-pattern',
        metavar='REGEX',
        help='''
        A regular expression specifying which Git tags to match.

        This pattern is used as a regex search for any incoming pushes. For
        example, --tag-pattern=foo will match "foo", "foobar", and "barfoo".
        Events on a tag that does not match will be ignored.

        The syntax of the regular expressions accepted is the syntax accepted by
        RE2 and described at https://github.com/google/re2/wiki/Syntax.
        ''')
    pr_preview = trigger_match.add_argument_group(
        help='Pull request preview deployment settings')
    pr_preview.add_argument(
        '--pull-request-preview',
        help='''
        Enables previewing your application for each pull request.

        This configures your application to deploy to a target cluster when
        a pull request is created or updated against a branch specified by the
        `--pull-request-pattern` argument. The application will be deployed
        to the namespace 'preview-[REPO_NAME]-[PR_NUMBER]'. This namespace will
        be deleted after a number of days specified by the `--preview-expiry`
        argument.

        The deployed preview application will still exist even after the pull
        request is merged or closed. The preview application will eventually get
        cleaned up by a Cloud Scheduler job after the namespace expires. You can
        also delete the namespace manually.
        ''',
        action='store_true',
        required=True
    )
    pr_preview.add_argument(
        '--preview-expiry',
        type=int,
        default=3,
        help='''
        Number of days before a pull request preview deployment's namespace is
        considered to be expired. An expired namespace will eventually be
        deleted. Defaults to 3 days.
        '''
    )
    pr_preview.add_argument(
        '--pull-request-pattern',
        metavar='REGEX',
        help="""
        A regular expression specifying which base Git branch to match for
        pull request events.

        This pattern is used as a regex search for the base branch (the branch
        you are trying to merge into) for pull request updates. For example,
        --pull-request-pattern=foo will match "foo", "foobar", and "barfoo".

        The syntax of the regular expressions accepted is the syntax accepted by
        RE2 and described at https://github.com/google/re2/wiki/Syntax.
        """,
        required=True
    )
    pr_preview.add_argument(
        '--comment-control',
        help="Require a repo collaborator to add '/gcbrun' as a comment in the "
        'pull request in order to run the build.',
        action='store_true'
    )
    parser.add_argument(
        '--gcs-config-staging-dir',
        help="""
        Path to the Google Cloud Storage subdirectory into which to copy the
        configs (suggested base and expanded Kubernetes YAML files) that are
        used to stage and deploy your app. If the bucket in this path doesn't
        exist, Cloud Build creates it.

        If this field is not set, the configs are written to
        ```gs://[PROJECT_ID]_cloudbuild/deploy/config```.
        """)
    parser.add_argument(
        '--app-name',
        help='If specified, the following label is added to the Kubernetes '
        "manifests: 'app.kubernetes.io/name: APP_NAME'. Defaults to the "
        'repository name provided by `--repo-name`.')
    parser.add_argument(
        '--cluster',
        help='Name of the target cluster to deploy to.',
        required=True)
    parser.add_argument(
        '--location',
        help='Region or zone of the target cluster to deploy to.',
        required=True)
    parser.add_argument(
        '--namespace',
        help='Namespace of the target cluster to deploy to. If this field is '
        "not set, the 'default' namespace is used.")
    parser.add_argument(
        '--config',
        help="""
        Path to the Kubernetes YAML, or directory containing multiple
        Kubernetes YAML files, used to deploy the container image. The path is
        relative to the repository root. The files must reference the provided
        container image or tag.

        If this field is not set, a default Deployment config and Horizontal
        Pod Autoscaler config are used to deploy the image.
        """)
    parser.add_argument(
        '--expose',
        type=int,
        help='Port that the deployed application listens on. If set, a '
        "Kubernetes Service of type 'LoadBalancer' is created with a "
        'single TCP port mapping that exposes this port.')
    parser.add_argument(
        '--timeout',
        help='Maximum time a build is run before it times out. For example, '
        '"2h15m5s" is two hours, fifteen minutes, and five seconds. If you '
        'do not specify a unit, seconds is assumed. Overrides the default '
        'builds/timeout property value for this command invocation.',
        action=actions.StoreProperty(properties.VALUES.builds.timeout))

  def Run(self, args):
    """This is what gets called when the user runs this command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      Some value that we want to have printed later.
    """

    if args.pull_request_preview:
      if args.repo_type != 'github':
        raise c_exceptions.InvalidArgumentException(
            '--repo-type',
            "Repo type must be 'github' to configure pull request previewing.")
      if args.namespace:
        raise c_exceptions.InvalidArgumentException(
            '--namespace',
            'Namespace must not be provided to configure pull request '
            'previewing. --namespace must only be provided when configuring '
            'automated deployments with the --branch-pattern or --tag-pattern '
            'flags.')
      if args.preview_expiry <= 0:
        raise c_exceptions.InvalidArgumentException(
            '--preview-expiry',
            'Preview expiry must be > 0.')

    # Determine image based on repo type
    image = None

    # Determine github app or csr
    github_repo_name = None
    github_repo_owner = None
    csr_repo_name = None

    project = properties.VALUES.core.project.Get(required=True)

    if args.repo_type == 'github':
      if not args.repo_owner:
        raise c_exceptions.RequiredArgumentException(
            '--repo-owner',
            'Repo owner is required for --repo-type=github.')
      image = 'gcr.io/{}/github.com/{}/{}:$COMMIT_SHA'.format(
          project, args.repo_owner, args.repo_name)
      github_repo_name = args.repo_name
      github_repo_owner = args.repo_owner
      # We do not have to verify that this repo exists because the request to
      # create the BuildTrigger will fail with the appropriate message asking
      # the user to connect their repo, if the repo is not found.

    elif args.repo_type == 'csr':
      if args.repo_owner:
        raise c_exceptions.InvalidArgumentException(
            '--repo-owner',
            'Repo owner must not be provided for --repo-type=csr.')
      image = 'gcr.io/{}/{}:$COMMIT_SHA'.format(project, args.repo_name)
      csr_repo_name = args.repo_name
      self._VerifyCSRRepoExists(csr_repo_name)

    elif args.repo_type == 'bitbucket_mirrored':
      if not args.repo_owner:
        raise c_exceptions.RequiredArgumentException(
            '--repo-owner',
            'Repo owner is required for --repo-type=bitbucket_mirrored.')
      image = 'gcr.io/{}/bitbucket.org/{}/{}:$COMMIT_SHA'.format(
          project, args.repo_owner, args.repo_name)
      csr_repo_name = 'bitbucket_{}_{}'.format(args.repo_owner, args.repo_name)
      self._VerifyBitbucketCSRRepoExists(
          csr_repo_name, args.repo_owner, args.repo_name)

    elif args.repo_type == 'github_mirrored':
      if not args.repo_owner:
        raise c_exceptions.RequiredArgumentException(
            '--repo-owner',
            'Repo owner is required for --repo-type=github_mirrored.')
      image = 'gcr.io/{}/github.com/{}/{}:$COMMIT_SHA'.format(
          project, args.repo_owner, args.repo_name)
      csr_repo_name = 'github_{}_{}'.format(args.repo_owner, args.repo_name)
      self._VerifyGitHubCSRRepoExists(
          csr_repo_name, args.repo_owner, args.repo_name)

    self._VerifyClusterExists(args.cluster, args.location)

    # Determine app_name
    if args.app_name:
      app_name = args.app_name
    else:
      app_name = args.repo_name

    # Determine gcs_config_staging_dir_bucket, gcs_config_staging_dir_object
    if args.gcs_config_staging_dir is None:
      gcs_config_staging_dir_bucket = \
        staging_bucket_util.GetDefaultStagingBucket()
      gcs_config_staging_dir_object = 'deploy/config'
    else:
      try:
        gcs_config_staging_dir_ref = resources.REGISTRY.Parse(
            args.gcs_config_staging_dir, collection='storage.objects')
        gcs_config_staging_dir_object = gcs_config_staging_dir_ref.object
      except resources.WrongResourceCollectionException:
        gcs_config_staging_dir_ref = resources.REGISTRY.Parse(
            args.gcs_config_staging_dir, collection='storage.buckets')
        gcs_config_staging_dir_object = None
      gcs_config_staging_dir_bucket = gcs_config_staging_dir_ref.bucket

    gcs_client = storage_api.StorageClient()
    try:
      gcs_client.CreateBucketIfNotExists(
          gcs_config_staging_dir_bucket,
          check_ownership=args.gcs_config_staging_dir is None)
    except storage_api.BucketInWrongProjectError:
      # If we're using the default bucket but it already exists in a different
      # project, then it could belong to a malicious attacker (b/33046325).
      raise c_exceptions.RequiredArgumentException(
          '--gcs-config-staging-dir',
          'A bucket with name {} already exists and is owned by '
          'another project. Specify a bucket using '
          '--gcs-config-staging-dir.'.format(gcs_config_staging_dir_bucket))

    if gcs_config_staging_dir_object:
      gcs_config_staging_path = '{}/{}'.format(
          gcs_config_staging_dir_bucket, gcs_config_staging_dir_object)
    else:
      gcs_config_staging_path = gcs_config_staging_dir_bucket

    if args.pull_request_preview:
      log.status.Print('Setting up previewing {} on pull requests.\n'.format(
          github_repo_name))
      self._ConfigurePRPreview(
          repo_owner=github_repo_owner,
          repo_name=github_repo_name,
          pull_request_pattern=args.pull_request_pattern,
          preview_expiry=args.preview_expiry,
          comment_control=args.comment_control,
          image=image,
          dockerfile_path=args.dockerfile,
          app_name=app_name,
          config_path=args.config,
          expose_port=args.expose,
          gcs_config_staging_path=gcs_config_staging_path,
          cluster=args.cluster,
          location=args.location)
    else:
      log.status.Print('Setting up automated deployments for {}.\n'.format(
          args.repo_name))
      self._ConfigureGitPushBuildTrigger(
          repo_type=args.repo_type,
          csr_repo_name=csr_repo_name,
          github_repo_owner=github_repo_owner,
          github_repo_name=github_repo_name,
          branch_pattern=args.branch_pattern,
          tag_pattern=args.tag_pattern,
          image=image,
          dockerfile_path=args.dockerfile,
          app_name=app_name,
          config_path=args.config,
          namespace=args.namespace,
          expose_port=args.expose,
          gcs_config_staging_path=gcs_config_staging_path,
          cluster=args.cluster,
          location=args.location)

  def _VerifyCSRRepoExists(self, csr_repo_name):
    try:
      csr_repo_ref = sourcerepo.ParseRepo(csr_repo_name)
      csr_repo = sourcerepo.Source().GetRepo(csr_repo_ref)
      if csr_repo.mirrorConfig:
        log.error("CSR repo '{}' has mirrorConfig.url of {}", csr_repo_name,
                  csr_repo.mirrorConfig.url)
        raise c_exceptions.InvalidArgumentException(
            '--repo-type',
            "Repo '{}' is found but is connected to {}. Specify the correct "
            'value for --repo-type, along with appropriate values for '
            '--repo-owner and --repo-name.'
            .format(csr_repo_name, csr_repo.mirrorConfig.url)
        )
    except HttpNotFoundError:
      raise c_exceptions.InvalidArgumentException(
          '--repo-name',
          "Repo '{}' is not found on CSR.".format(csr_repo_name))

  def _VerifyBitbucketCSRRepoExists(self, csr_repo_name, bitbucket_repo_owner,
                                    bitbucket_repo_name):
    try:
      csr_repo_ref = sourcerepo.ParseRepo(csr_repo_name)
      csr_repo = sourcerepo.Source().GetRepo(csr_repo_ref)
      if not csr_repo.mirrorConfig:
        raise c_exceptions.InvalidArgumentException(
            '--repo-type',
            "Repo '{}/{}' is found but the resolved repo name '{}' is a "
            'regular CSR repo. Reference it with --repo-type=csr and '
            '--repo-name={}.'
            .format(bitbucket_repo_owner, bitbucket_repo_name, csr_repo_name,
                    csr_repo_name))
      if (csr_repo.mirrorConfig and
          not csr_repo.mirrorConfig.url.startswith('https://bitbucket.org/')):
        log.error("CSR repo '{}' has mirrorConfig.url of {}", csr_repo_name,
                  csr_repo.mirrorConfig.url)
        raise c_exceptions.InvalidArgumentException(
            '--repo-type',
            "Repo '{}/{}' is found but the resolved repo name '{}' is not "
            'connected to a Bitbucket repo. Specify the correct value for '
            '--repo-type.'
            .format(bitbucket_repo_owner, bitbucket_repo_name, csr_repo_name))
    except HttpNotFoundError:
      raise SourceRepoNotConnectedException(
          '--repo-name',
          "Bitbucket repo '{}/{}' is not connected to CSR.".format(
              bitbucket_repo_owner, bitbucket_repo_name))

  def _VerifyGitHubCSRRepoExists(self, csr_repo_name, github_repo_owner,
                                 github_repo_name):
    try:
      csr_repo_ref = sourcerepo.ParseRepo(csr_repo_name)
      csr_repo = sourcerepo.Source().GetRepo(csr_repo_ref)
      if not csr_repo.mirrorConfig:
        raise c_exceptions.InvalidArgumentException(
            '--repo-type',
            "Repo '{}/{}' is found but the resolved repo name '{}' is a "
            'regular CSR repo. Reference it with --repo-type=csr and '
            '--repo-name={}.'
            .format(github_repo_owner, github_repo_name, csr_repo_name,
                    csr_repo_name))
      if (csr_repo.mirrorConfig and
          not csr_repo.mirrorConfig.url.startswith('https://github.com/')):
        log.error("CSR repo '{}' has mirrorConfig.url of {}", csr_repo_name,
                  csr_repo.mirrorConfig.url)
        raise c_exceptions.InvalidArgumentException(
            '--repo-type',
            "Repo '{}/{}' is found but the resolved repo name '{}' is not "
            'connected to a GitHub repo. Specify the correct value for '
            '--repo-type.'
            .format(github_repo_owner, github_repo_name, csr_repo_name))
    except HttpNotFoundError:
      raise SourceRepoNotConnectedException(
          '--repo-name',
          "GitHub repo '{}/{}' is not connected to CSR.".format(
              github_repo_owner, github_repo_name))

  def _VerifyClusterExists(self, cluster, location):
    client = apis.GetClientInstance('container', 'v1')
    messages = apis.GetMessagesModule('container', 'v1')
    project = properties.VALUES.core.project.Get(required=True)
    try:
      cluster_res = client.projects_locations_clusters.Get(
          messages.ContainerProjectsLocationsClustersGetRequest(
              name='projects/{project}/locations/{location}/clusters/{cluster}'
              .format(
                  project=project,
                  location=location,
                  cluster=cluster
              )))
    except HttpNotFoundError:
      raise c_exceptions.InvalidArgumentException(
          '--cluster',
          "No cluster '{cluster}' in location '{location}' in project {project}.\n\n"
          'Visit https://console.cloud.google.com/kubernetes/list?project={project} '
          'to create a cluster.'.format(
              cluster=cluster,
              location=location,
              project=project
          ))
    if cluster_res.status != messages.Cluster.StatusValueValuesEnum.RUNNING:
      raise core_exceptions.Error(
          'Cluster was found but status is not RUNNING. Status is {}.'
          .format(cluster_res.status))

  def _GetTriggerIfExists(self, client, messages, project, trigger_name):
    """Returns a BuildTrigger if one with the given name exists in a project.

    Args:
      client: Client used to make calls to Cloud Build API.
      messages: Cloud Build messages module. This is the value returned from
        cloudbuild_util.GetMessagesModule().
      project: Project of BuildTrigger to check existence.
      trigger_name: Name of BuildTrigger to check existence.

    Returns:
      A BuildTrigger with the given trigger_name if it exists, else None.
    """
    try:
      return client.projects_triggers.Get(
          messages.CloudbuildProjectsTriggersGetRequest(
              projectId=project,
              triggerId=trigger_name))  # Undocumented, but this field can be ID
                                       # or name.
    except HttpNotFoundError:
      return None

  def _UpsertBuildTrigger(self, build_trigger, add_gcb_trigger_id):
    """Creates a BuildTrigger using the CloudBuild API if it doesn't exist, else updates it.

    A BuildTrigger "exists" if one with the same name already exists in the
    project.

    Args:
      build_trigger: Config of BuildTrigger to create.
      add_gcb_trigger_id: If True, adds the gcb-trigger-id=<trigger-id>
        annotation to the deployed Kubernetes objects. The annotation must be
        added to an existing trigger because the trigger-id is only known after
        the trigger is created.

    Returns:
      The upserted trigger.
    """
    client = cloudbuild_util.GetClientInstance()
    messages = cloudbuild_util.GetMessagesModule()
    project = properties.VALUES.core.project.Get(required=True)

    # Check if trigger with this name already exists.
    existing = self._GetTriggerIfExists(
        client, messages, project, build_trigger.name)
    if existing:
      trigger_id = existing.id

      if add_gcb_trigger_id:
        # Use the existing trigger's id to patch the trigger object we created
        # to add gcb-trigger-id=<trigger-id> annotation for its deployed
        # resources.
        build_util.AddAnnotationToPrepareDeployStep(
            build_trigger, 'gcb-trigger-id', trigger_id)

      upserted_build_trigger = client.projects_triggers.Patch(
          messages.CloudbuildProjectsTriggersPatchRequest(
              buildTrigger=build_trigger,
              projectId=project,
              triggerId=trigger_id
          )
      )
      log.debug('updated existing BuildTrigger: '
                + six.text_type(upserted_build_trigger))

    else:
      upserted_build_trigger = client.projects_triggers.Create(
          messages.CloudbuildProjectsTriggersCreateRequest(
              buildTrigger=build_trigger, projectId=project))
      log.debug('created BuildTrigger: '
                + six.text_type(upserted_build_trigger))

      trigger_id = upserted_build_trigger.id
      if add_gcb_trigger_id:
        # Since <trigger-id> is only known after a BuildTrigger is created, we
        # must patch the newly created trigger to add
        # gcb-trigger-id=<trigger-id> annotation for its deployed resources.
        build_util.AddAnnotationToPrepareDeployStep(
            upserted_build_trigger, 'gcb-trigger-id', trigger_id)
        upserted_build_trigger = client.projects_triggers.Patch(
            messages.CloudbuildProjectsTriggersPatchRequest(
                buildTrigger=upserted_build_trigger,
                projectId=project,
                triggerId=trigger_id
            )
        )
        log.debug('updated BuildTrigger with gcb-trigger-id annotation: '
                  + six.text_type(upserted_build_trigger))

    # Log trigger full name
    build_trigger_ref = resources.REGISTRY.Parse(
        None,
        collection='cloudbuild.projects.triggers',
        api_version='v1',
        params={
            'projectId': project,
            'triggerId': trigger_id,
        })

    if existing:
      log.UpdatedResource(build_trigger_ref)
    else:
      log.CreatedResource(build_trigger_ref)

    return upserted_build_trigger

  def _ConfigureGitPushBuildTrigger(
      self, repo_type, csr_repo_name, github_repo_owner, github_repo_name,
      branch_pattern, tag_pattern, image, dockerfile_path, app_name,
      config_path, namespace, expose_port, gcs_config_staging_path, cluster,
      location):

    # Generate deterministic trigger name
    if csr_repo_name:
      full_repo_name = csr_repo_name
    else:
      full_repo_name = github_repo_owner + '-' + github_repo_name

    name = self._FixBuildTriggerName(self._GenerateResourceName(
        function_code='gp',  # Git Push
        repo_type=repo_type,
        full_repo_name=full_repo_name,
        branch_pattern=branch_pattern,
        tag_pattern=tag_pattern))

    if branch_pattern:
      description = 'Build and deploy on push to "{}"'.format(branch_pattern)
    elif tag_pattern:
      description = 'Build and deploy on "{}" tag'.format(tag_pattern)

    build_trigger = build_util.CreateGitPushBuildTrigger(
        cloudbuild_util.GetMessagesModule(),
        name=name,
        description=description,
        build_timeout=properties.VALUES.builds.timeout.Get(),
        csr_repo_name=csr_repo_name,
        github_repo_owner=github_repo_owner,
        github_repo_name=github_repo_name,
        branch_pattern=branch_pattern,
        tag_pattern=tag_pattern,
        image=image,
        dockerfile_path=dockerfile_path,
        app_name=app_name,
        config_path=config_path,
        namespace=namespace,
        expose_port=expose_port,
        gcs_config_staging_path=gcs_config_staging_path,
        cluster=cluster,
        location=location,
        build_tags=[app_name],
        build_trigger_tags=[app_name],
    )

    log.status.Print('Upserting Cloud Build trigger to build and deploy your '
                     'application.')
    upserted_trigger = self._UpsertBuildTrigger(build_trigger, True)
    project = properties.VALUES.core.project.Get(required=True)

    log.status.Print(
        '\nSuccessfully created the Cloud Build trigger to build and deploy '
        'your application.\n\n'
        'Visit https://console.cloud.google.com/cloud-build/triggers/edit/{trigger_id}?project={project} '
        'to view the trigger.\n\n'
        'You can visit https://console.cloud.google.com/cloud-build/triggers?project={project} '
        'to view all Cloud Build triggers.'.format(
            trigger_id=upserted_trigger.id, project=project)
    )

  def _ConfigurePRPreview(
      self, repo_owner, repo_name, pull_request_pattern, preview_expiry,
      comment_control, image, dockerfile_path, app_name, config_path,
      expose_port, gcs_config_staging_path, cluster, location):
    """Configures previewing the application for each pull request.

    PR previewing is only supported for GitHub repos.

    This creates three resources:
    * A BuildTrigger that builds, publishes, and deploys the application to
      namespace 'preview-[REPO_NAME]-[PR_NUMBER]. The deployed namespace has an
      expiry time, which indicates when it is considered to be expired.
    * A BuildTrigger that cleans up expired namespaces of this application.
    * A CloudScheduler Job that executes the BuildTrigger every day. This is
      needed because BuildTriggers can't run on a Cron by themselves.

    Args:
      repo_owner: Owner of repo to be deployed.
      repo_name: Name of repo to be deployed.
      pull_request_pattern: Regex value of branch to trigger on.
      preview_expiry: How long, in days, a preview namespace can exist before it
        is expired.
      comment_control: Whether or not a user must comment /gcbrun to trigger
        the deployment build.
      image: The image that will be built and deployed. The image can include a
        tag or digest.
      dockerfile_path: Path to the source repository's Dockerfile, relative to
        the source repository's root directory.
      app_name: Application name, which is set as a label to deployed objects.
      config_path: Path to the source repository's Kubernetes configs, relative
        to the source repository's root directory.
      expose_port: Port that the deployed application listens on.
      gcs_config_staging_path: Path to a GCS subdirectory to copy application
        configs to.
      cluster: Name of target cluster to deploy to.
      location: Zone/region of target cluster to deploy to.
    """
    # Attempt to get scheduler location before creating any resources so we can
    # fail early if we fail to get the location due to the user's App Engine
    # app not existing, since the scheduler job must be in the same region:
    # https://cloud.google.com/scheduler/docs/
    scheduler_location = self._GetSchedulerJobLocation()

    pr_preview_trigger = self._ConfigurePRPreviewBuildTrigger(
        repo_owner=repo_owner,
        repo_name=repo_name,
        pull_request_pattern=pull_request_pattern,
        preview_expiry=preview_expiry,
        comment_control=comment_control,
        image=image,
        dockerfile_path=dockerfile_path,
        app_name=app_name,
        config_path=config_path,
        expose_port=expose_port,
        gcs_config_staging_path=gcs_config_staging_path,
        cluster=cluster,
        location=location
    )
    clean_preview_trigger = self._ConfigureCleanPreviewBuildTrigger(
        repo_owner=repo_owner,
        repo_name=repo_name,
        pull_request_pattern=pull_request_pattern,
        cluster=cluster,
        location=location,
        app_name=app_name
    )
    # TODO(b/141294571): Once cron triggers feature is done, use cron trigger
    #   instead of Scheduler job.
    clean_preview_job = self._ConfigureCleanPreviewSchedulerJob(
        repo_owner=repo_owner,
        repo_name=repo_name,
        pull_request_pattern=pull_request_pattern,
        clean_preview_trigger_id=clean_preview_trigger.id,
        scheduler_location=scheduler_location)

    # We add scheduler job location and name as tags to the clean preview
    # trigger to track it as created by us.
    job_location = clean_preview_job.name.split('/')[-3]
    job_id = clean_preview_job.name.split('/')[-1]
    self._UpdateBuildTriggerWithSchedulerJobTags(
        clean_preview_trigger, job_location, job_id)

    log.status.Print(
        '\nSuccessfully created resources for previewing pull requests of your '
        'application.\n\n'
        'Visit https://console.cloud.google.com/cloud-build/triggers/edit/{pr_preview_trigger_id}?project={project} '
        'to view the Cloud Build trigger that deploys your application on pull '
        'request open/update.\n\n'
        'Visit https://console.cloud.google.com/cloud-build/triggers/edit/{clean_preview_trigger_id}?project={project} '
        'to view the Cloud Build trigger that cleans up expired preview '
        'deployments.\n\n'
        'Visit https://console.cloud.google.com/cloudscheduler/jobs/edit/{location}/{job}?project={project} '
        'to view the Cloud Scheduler job that periodically executes the '
        'trigger to clean up expired preview deployments.\n\n'
        'WARNING: The deletion of expired preview deployments requires a Cloud '
        'Scheduler job that runs a Cloud Build trigger every day. Pause '
        "this job if you don't want to it to run anymore."
        .format(
            pr_preview_trigger_id=pr_preview_trigger.id,
            clean_preview_trigger_id=clean_preview_trigger.id,
            project=properties.VALUES.core.project.Get(required=True),
            location=job_location,
            job=job_id
        )
    )

  def _UpdateBuildTriggerWithSchedulerJobTags(
      self, build_trigger, job_location, job_id):
    job_location_tag = 'cloudscheduler-job-location_' + job_location
    job_id_tag = 'cloudscheduler-job-id_' + job_id

    build_trigger_tags = [x for x in build_trigger.tags
                          if not x.startswith('cloudscheduler-job-')]
    build_trigger_tags.append(job_location_tag)
    build_trigger_tags.append(job_id_tag)
    build_trigger.tags = build_trigger_tags

    build_tags = [x for x in build_trigger.build.tags
                  if not x.startswith('cloudscheduler-job-')]
    build_tags.append(job_location_tag)
    build_tags.append(job_id_tag)
    build_trigger.build.tags = build_tags

    client = cloudbuild_util.GetClientInstance()
    messages = cloudbuild_util.GetMessagesModule()
    project = properties.VALUES.core.project.Get(required=True)

    updated_trigger = client.projects_triggers.Patch(
        messages.CloudbuildProjectsTriggersPatchRequest(
            buildTrigger=build_trigger,
            projectId=project,
            triggerId=build_trigger.id
        )
    )

    log.debug('added job id to trigger: ' + six.text_type(updated_trigger))

  def _GenerateResourceName(
      self, function_code, repo_type, full_repo_name,
      branch_pattern=None, tag_pattern=None):
    """Generate a short, deterministic resource name based on parameters that should be unique.

    Args:
      function_code: A two-character code describing what function the resource
        has.
      repo_type: A two-character code describing the repo type.
      full_repo_name: Deterministicly generated repo name, including owner
        available.
      branch_pattern: Branch pattern to match. Only one of branch_pattern or
        tag_pattern should be provided. They can also both be omitted.
      tag_pattern: Tag pattern to match. Only one of branch_pattern or
        tag_pattern should be provided. They can also both be omitted.

    Returns:
      Deterministicly generated resource name.
    """

    repo_type_code = _REPO_TYPE_CODES[repo_type]
    if branch_pattern:
      return '{}{}b-{}-{}'.format(
          function_code, repo_type_code, full_repo_name, branch_pattern)
    elif tag_pattern:
      return '{}{}t-{}-{}'.format(
          function_code, repo_type_code, full_repo_name, tag_pattern)
    else:
      return '{}{}b-{}'.format(
          function_code, repo_type_code, full_repo_name)

  def _ConfigurePRPreviewBuildTrigger(
      self, repo_owner, repo_name, pull_request_pattern, preview_expiry,
      comment_control, image, dockerfile_path, app_name, config_path,
      expose_port, gcs_config_staging_path, cluster, location):

    # Generate deterministic trigger name
    name = self._FixBuildTriggerName(self._GenerateResourceName(
        function_code='pp',  # Pr Preview
        repo_type='github',  # Only supports github for now.
        full_repo_name=repo_owner + '-' + repo_name))
    description = \
      'Build and deploy on PR create/update against "{}"'.format(
          pull_request_pattern)

    build_trigger = build_util.CreatePRPreviewBuildTrigger(
        messages=cloudbuild_util.GetMessagesModule(),
        name=name,
        description=description,
        build_timeout=properties.VALUES.builds.timeout.Get(),
        github_repo_owner=repo_owner,
        github_repo_name=repo_name,
        pr_pattern=pull_request_pattern,
        preview_expiry_days=preview_expiry,
        comment_control=comment_control,
        image=image,
        dockerfile_path=dockerfile_path,
        app_name=app_name,
        config_path=config_path,
        expose_port=expose_port,
        gcs_config_staging_path=gcs_config_staging_path,
        cluster=cluster,
        location=location,
        build_tags=[app_name],
        build_trigger_tags=[app_name]
    )

    log.status.Print('Upserting Cloud Build trigger to build and deploy your '
                     'application on PR open/update for previewing.')
    return self._UpsertBuildTrigger(build_trigger, True)

  def _ConfigureCleanPreviewBuildTrigger(
      self, repo_name, repo_owner, pull_request_pattern, cluster, location,
      app_name):

    # Generate deterministic trigger name
    name = self._FixBuildTriggerName(self._GenerateResourceName(
        function_code='cp',  # Clean Preview
        repo_type='github',  # Only supports github for now.
        full_repo_name=repo_owner + '-' + repo_name))
    description = \
        'Clean expired preview deployments for PRs against "{}"'.format(
            pull_request_pattern)

    build_trigger = build_util.CreateCleanPreviewBuildTrigger(
        messages=cloudbuild_util.GetMessagesModule(),
        name=name,
        description=description,
        github_repo_owner=repo_owner,
        github_repo_name=repo_name,
        cluster=cluster,
        location=location,
        build_tags=[app_name],
        build_trigger_tags=[app_name]
    )

    log.status.Print('Upserting Cloud Build trigger to clean expired preview '
                     'deployments of your application.')
    return self._UpsertBuildTrigger(build_trigger, False)

  def _GetSchedulerJobLocation(self):
    messages = apis.GetMessagesModule('cloudscheduler', 'v1')
    client = apis.GetClientInstance('cloudscheduler', 'v1')
    project = properties.VALUES.core.project.Get(required=True)

    try:
      locations_res = client.projects_locations.List(
          messages.CloudschedulerProjectsLocationsListRequest(
              name='projects/' + project))
    except HttpNotFoundError:
      raise core_exceptions.Error(
          'You must create an App Engine application in your project to use '
          'Cloud Scheduler. Visit '
          'https://console.developers.google.com/appengine?project={} to '
          'add an App Engine application.'.format(project))

    return locations_res.locations[0].labels.additionalProperties[0].value

  def _ConfigureCleanPreviewSchedulerJob(
      self, repo_owner, repo_name, pull_request_pattern,
      clean_preview_trigger_id, scheduler_location):

    log.status.Print('Upserting Cloud Scheduler to run Cloud Build trigger to '
                     'clean expired preview deployments of your application.')

    messages = apis.GetMessagesModule('cloudscheduler', 'v1')
    client = apis.GetClientInstance('cloudscheduler', 'v1')
    project = properties.VALUES.core.project.Get(required=True)
    service_account_email = project + '@appspot.gserviceaccount.com'

    # Generate deterministic scheduler job name (id)
    job_id = self._FixSchedulerName(self._GenerateResourceName(
        function_code='cp',  # Clean Preview
        repo_type='github',  # Only supports github for now.
        full_repo_name=repo_owner + '-' + repo_name))

    name = 'projects/{}/locations/{}/jobs/{}'.format(
        project, scheduler_location, job_id)

    job = messages.Job(
        name=name,
        description='Every day, run trigger to clean expired preview '
                    'deployments for PRs against "{}" in {}/{}'.format(
                        pull_request_pattern, repo_owner, repo_name),
        schedule=_CLEAN_PREVIEW_SCHEDULE,
        timeZone='UTC',
        httpTarget=messages.HttpTarget(
            uri='https://cloudbuild.googleapis.com/v1/projects/{}/triggers/{}:run'
            .format(project, clean_preview_trigger_id),
            httpMethod=messages.HttpTarget.HttpMethodValueValuesEnum.POST,
            body=bytes(
                # We don't actually use the branchName value but it has to be
                # set to an existing branch, so set it to master.
                '{{"projectId":"{}","repoName":"{}","branchName":"master"}}'
                .format(project, repo_name)
                .encode('utf-8')),
            oauthToken=messages.OAuthToken(
                serviceAccountEmail=service_account_email
            )
        )
    )

    existing = None
    try:
      existing = client.projects_locations_jobs.Get(
          messages.CloudschedulerProjectsLocationsJobsGetRequest(name=name))

      upserted_job = client.projects_locations_jobs.Patch(
          messages.CloudschedulerProjectsLocationsJobsPatchRequest(
              name=name,
              job=job))
      log.debug('updated existing CloudScheduler job: '
                + six.text_type(upserted_job))

    except HttpNotFoundError:
      upserted_job = client.projects_locations_jobs.Create(
          messages.CloudschedulerProjectsLocationsJobsCreateRequest(
              parent='projects/{}/locations/{}'.format(
                  project, scheduler_location),
              job=job))
      log.debug('created CloudScheduler job: ' + six.text_type(upserted_job))

    job_id = upserted_job.name.split('/')[-1]
    job_ref = resources.REGISTRY.Parse(
        None,
        collection='cloudscheduler.projects.locations.jobs',
        api_version='v1',
        params={
            'projectsId': project,
            'locationsId': scheduler_location,
            'jobsId': job_id,
        })

    if existing:
      log.UpdatedResource(job_ref)
    else:
      log.CreatedResource(job_ref)

    return upserted_job

  def _FixBuildTriggerName(self, name):
    """Fixes a BuildTrigger name to match the allowed format.

    Args:
      name: Name must start with an alpha-numberic character, since this method
        will not check that condition.

    Returns:
      Fixed name.
    """
    # Name pattern must match
    # ^[a-zA-Z0-9]$|^[a-zA-Z0-9][a-zA-Z0-9-]{0,62}[a-zA-Z0-9]$, where the max
    # length is 64 characters.
    if len(name) > 64:
      name = name[:64]
    name = re.sub('[^a-zA-Z0-9-]', '-', name)
    if name.endswith('-'):  # Cannot trail in a '-' character.
      name = name[:-1] + '0'
    # Assume the name starts with alpha-numeric character, so we will not check.

    return name

  def _FixSchedulerName(self, name):
    """Fixes a Scheduler Job's name to match the allowed format.

    Args:
      name: Name to fix.

    Returns:
      Fixed name.
    """
    # Name pattern must match "[a-zA-Z\d_-]{1,500}".
    if len(name) > 500:
      name = name[:500]
    name = re.sub('[^a-zA-Z0-9_-]', '-', name)

    return name



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Callable

from google.api_core.exceptions import FailedPrecondition
from google.pubsub_v1 import PubsubMessage


class NackHandler(ABC):
    """
    A NackHandler handles calls to the nack() method which is not expressible in Pub/Sub Lite.
    """

    @abstractmethod
    def on_nack(self, message: PubsubMessage, ack: Callable[[], None]):
        """Handle a negative acknowledgement. ack must eventually be called.

        This method will be called on an event loop and should not block.

        Args:
          message: The nacked message.
          ack: A callable to acknowledge the underlying message. This must eventually be called.

        Raises:
          GoogleAPICallError: To fail the client if raised inline.
        """
        pass


class DefaultNackHandler(NackHandler):
    def on_nack(self, message: PubsubMessage, ack: Callable[[], None]):
        raise FailedPrecondition(
            "You may not nack messages by default when using a PubSub Lite client. See NackHandler for how to customize"
            " this."
        )



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import OrderedDict
from distutils import util
import os
import re
from typing import Callable, Dict, Optional, Sequence, Tuple, Type, Union
import pkg_resources

from google.api_core import client_options as client_options_lib  # type: ignore
from google.api_core import exceptions  # type: ignore
from google.api_core import gapic_v1  # type: ignore
from google.api_core import retry as retries  # type: ignore
from google.auth import credentials  # type: ignore
from google.auth.transport import mtls  # type: ignore
from google.auth.transport.grpc import SslCredentials  # type: ignore
from google.auth.exceptions import MutualTLSChannelError  # type: ignore
from google.oauth2 import service_account  # type: ignore

from google.iam.v1 import iam_policy_pb2 as iam_policy  # type: ignore
from google.iam.v1 import policy_pb2 as policy  # type: ignore
from google.pubsub_v1.services.schema_service import pagers
from google.pubsub_v1.types import schema
from google.pubsub_v1.types import schema as gp_schema

from .transports.base import SchemaServiceTransport, DEFAULT_CLIENT_INFO
from .transports.grpc import SchemaServiceGrpcTransport
from .transports.grpc_asyncio import SchemaServiceGrpcAsyncIOTransport


class SchemaServiceClientMeta(type):
    """Metaclass for the SchemaService client.

    This provides class-level methods for building and retrieving
    support objects (e.g. transport) without polluting the client instance
    objects.
    """

    _transport_registry = OrderedDict()  # type: Dict[str, Type[SchemaServiceTransport]]
    _transport_registry["grpc"] = SchemaServiceGrpcTransport
    _transport_registry["grpc_asyncio"] = SchemaServiceGrpcAsyncIOTransport

    def get_transport_class(cls, label: str = None,) -> Type[SchemaServiceTransport]:
        """Return an appropriate transport class.

        Args:
            label: The name of the desired transport. If none is
                provided, then the first transport in the registry is used.

        Returns:
            The transport class to use.
        """
        # If a specific transport is requested, return that one.
        if label:
            return cls._transport_registry[label]

        # No transport is requested; return the default (that is, the first one
        # in the dictionary).
        return next(iter(cls._transport_registry.values()))


class SchemaServiceClient(metaclass=SchemaServiceClientMeta):
    """ Service for doing schema-related operations.
    EXPERIMENTAL: The Schema service is in development and may not
    work yet.
    """

    @staticmethod
    def _get_default_mtls_endpoint(api_endpoint):
        """Convert api endpoint to mTLS endpoint.
        Convert "*.sandbox.googleapis.com" and "*.googleapis.com" to
        "*.mtls.sandbox.googleapis.com" and "*.mtls.googleapis.com" respectively.
        Args:
            api_endpoint (Optional[str]): the api endpoint to convert.
        Returns:
            str: converted mTLS api endpoint.
        """
        if not api_endpoint:
            return api_endpoint

        mtls_endpoint_re = re.compile(
            r"(?P<name>[^.]+)(?P<mtls>\.mtls)?(?P<sandbox>\.sandbox)?(?P<googledomain>\.googleapis\.com)?"
        )

        m = mtls_endpoint_re.match(api_endpoint)
        name, mtls, sandbox, googledomain = m.groups()
        if mtls or not googledomain:
            return api_endpoint

        if sandbox:
            return api_endpoint.replace(
                "sandbox.googleapis.com", "mtls.sandbox.googleapis.com"
            )

        return api_endpoint.replace(".googleapis.com", ".mtls.googleapis.com")

    DEFAULT_ENDPOINT = "pubsub.googleapis.com"
    DEFAULT_MTLS_ENDPOINT = _get_default_mtls_endpoint.__func__(  # type: ignore
        DEFAULT_ENDPOINT
    )

    @classmethod
    def from_service_account_file(cls, filename: str, *args, **kwargs):
        """Creates an instance of this client using the provided credentials
        file.

        Args:
            filename (str): The path to the service account private key json
                file.
            args: Additional arguments to pass to the constructor.
            kwargs: Additional arguments to pass to the constructor.

        Returns:
            {@api.name}: The constructed client.
        """
        credentials = service_account.Credentials.from_service_account_file(filename)
        kwargs["credentials"] = credentials
        return cls(*args, **kwargs)

    from_service_account_json = from_service_account_file

    @property
    def transport(self) -> SchemaServiceTransport:
        """Return the transport used by the client instance.

        Returns:
            SchemaServiceTransport: The transport used by the client instance.
        """
        return self._transport

    @staticmethod
    def schema_path(project: str, schema: str,) -> str:
        """Return a fully-qualified schema string."""
        return "projects/{project}/schemas/{schema}".format(
            project=project, schema=schema,
        )

    @staticmethod
    def parse_schema_path(path: str) -> Dict[str, str]:
        """Parse a schema path into its component segments."""
        m = re.match(r"^projects/(?P<project>.+?)/schemas/(?P<schema>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_billing_account_path(billing_account: str,) -> str:
        """Return a fully-qualified billing_account string."""
        return "billingAccounts/{billing_account}".format(
            billing_account=billing_account,
        )

    @staticmethod
    def parse_common_billing_account_path(path: str) -> Dict[str, str]:
        """Parse a billing_account path into its component segments."""
        m = re.match(r"^billingAccounts/(?P<billing_account>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_folder_path(folder: str,) -> str:
        """Return a fully-qualified folder string."""
        return "folders/{folder}".format(folder=folder,)

    @staticmethod
    def parse_common_folder_path(path: str) -> Dict[str, str]:
        """Parse a folder path into its component segments."""
        m = re.match(r"^folders/(?P<folder>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_organization_path(organization: str,) -> str:
        """Return a fully-qualified organization string."""
        return "organizations/{organization}".format(organization=organization,)

    @staticmethod
    def parse_common_organization_path(path: str) -> Dict[str, str]:
        """Parse a organization path into its component segments."""
        m = re.match(r"^organizations/(?P<organization>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_project_path(project: str,) -> str:
        """Return a fully-qualified project string."""
        return "projects/{project}".format(project=project,)

    @staticmethod
    def parse_common_project_path(path: str) -> Dict[str, str]:
        """Parse a project path into its component segments."""
        m = re.match(r"^projects/(?P<project>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_location_path(project: str, location: str,) -> str:
        """Return a fully-qualified location string."""
        return "projects/{project}/locations/{location}".format(
            project=project, location=location,
        )

    @staticmethod
    def parse_common_location_path(path: str) -> Dict[str, str]:
        """Parse a location path into its component segments."""
        m = re.match(r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)$", path)
        return m.groupdict() if m else {}

    def __init__(
        self,
        *,
        credentials: Optional[credentials.Credentials] = None,
        transport: Union[str, SchemaServiceTransport, None] = None,
        client_options: Optional[client_options_lib.ClientOptions] = None,
        client_info: gapic_v1.client_info.ClientInfo = DEFAULT_CLIENT_INFO,
    ) -> None:
        """Instantiate the schema service client.

        Args:
            credentials (Optional[google.auth.credentials.Credentials]): The
                authorization credentials to attach to requests. These
                credentials identify the application to the service; if none
                are specified, the client will attempt to ascertain the
                credentials from the environment.
            transport (Union[str, ~.SchemaServiceTransport]): The
                transport to use. If set to None, a transport is chosen
                automatically.
            client_options (client_options_lib.ClientOptions): Custom options for the
                client. It won't take effect if a ``transport`` instance is provided.
                (1) The ``api_endpoint`` property can be used to override the
                default endpoint provided by the client. GOOGLE_API_USE_MTLS_ENDPOINT
                environment variable can also be used to override the endpoint:
                "always" (always use the default mTLS endpoint), "never" (always
                use the default regular endpoint) and "auto" (auto switch to the
                default mTLS endpoint if client certificate is present, this is
                the default value). However, the ``api_endpoint`` property takes
                precedence if provided.
                (2) If GOOGLE_API_USE_CLIENT_CERTIFICATE environment variable
                is "true", then the ``client_cert_source`` property can be used
                to provide client certificate for mutual TLS transport. If
                not provided, the default SSL client certificate will be used if
                present. If GOOGLE_API_USE_CLIENT_CERTIFICATE is "false" or not
                set, no client certificate will be used.
            client_info (google.api_core.gapic_v1.client_info.ClientInfo):
                The client info used to send a user-agent string along with
                API requests. If ``None``, then default info will be used.
                Generally, you only need to set this if you're developing
                your own client library.

        Raises:
            google.auth.exceptions.MutualTLSChannelError: If mutual TLS transport
                creation failed for any reason.
        """
        if isinstance(client_options, dict):
            client_options = client_options_lib.from_dict(client_options)
        if client_options is None:
            client_options = client_options_lib.ClientOptions()

        # Create SSL credentials for mutual TLS if needed.
        use_client_cert = bool(
            util.strtobool(os.getenv("GOOGLE_API_USE_CLIENT_CERTIFICATE", "false"))
        )

        ssl_credentials = None
        is_mtls = False
        if use_client_cert:
            if client_options.client_cert_source:
                import grpc  # type: ignore

                cert, key = client_options.client_cert_source()
                ssl_credentials = grpc.ssl_channel_credentials(
                    certificate_chain=cert, private_key=key
                )
                is_mtls = True
            else:
                creds = SslCredentials()
                is_mtls = creds.is_mtls
                ssl_credentials = creds.ssl_credentials if is_mtls else None

        # Figure out which api endpoint to use.
        if client_options.api_endpoint is not None:
            api_endpoint = client_options.api_endpoint
        else:
            use_mtls_env = os.getenv("GOOGLE_API_USE_MTLS_ENDPOINT", "auto")
            if use_mtls_env == "never":
                api_endpoint = self.DEFAULT_ENDPOINT
            elif use_mtls_env == "always":
                api_endpoint = self.DEFAULT_MTLS_ENDPOINT
            elif use_mtls_env == "auto":
                api_endpoint = (
                    self.DEFAULT_MTLS_ENDPOINT if is_mtls else self.DEFAULT_ENDPOINT
                )
            else:
                raise MutualTLSChannelError(
                    "Unsupported GOOGLE_API_USE_MTLS_ENDPOINT value. Accepted values: never, auto, always"
                )

        # Save or instantiate the transport.
        # Ordinarily, we provide the transport, but allowing a custom transport
        # instance provides an extensibility point for unusual situations.
        if isinstance(transport, SchemaServiceTransport):
            # transport is a SchemaServiceTransport instance.
            if credentials or client_options.credentials_file:
                raise ValueError(
                    "When providing a transport instance, "
                    "provide its credentials directly."
                )
            if client_options.scopes:
                raise ValueError(
                    "When providing a transport instance, "
                    "provide its scopes directly."
                )
            self._transport = transport
        else:
            Transport = type(self).get_transport_class(transport)
            self._transport = Transport(
                credentials=credentials,
                credentials_file=client_options.credentials_file,
                host=api_endpoint,
                scopes=client_options.scopes,
                ssl_channel_credentials=ssl_credentials,
                quota_project_id=client_options.quota_project_id,
                client_info=client_info,
            )

    def create_schema(
        self,
        request: gp_schema.CreateSchemaRequest = None,
        *,
        parent: str = None,
        schema: gp_schema.Schema = None,
        schema_id: str = None,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> gp_schema.Schema:
        r"""Creates a schema.

        Args:
            request (:class:`~.gp_schema.CreateSchemaRequest`):
                The request object. Request for the CreateSchema method.
            parent (:class:`str`):
                Required. The name of the project in which to create the
                schema. Format is ``projects/{project-id}``.
                This corresponds to the ``parent`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            schema (:class:`~.gp_schema.Schema`):
                Required. The schema object to create.

                This schema's ``name`` parameter is ignored. The schema
                object returned by CreateSchema will have a ``name``
                made using the given ``parent`` and ``schema_id``.
                This corresponds to the ``schema`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            schema_id (:class:`str`):
                The ID to use for the schema, which will become the
                final component of the schema's resource name.

                See
                https://cloud.google.com/pubsub/docs/admin#resource_names
                for resource name constraints.
                This corresponds to the ``schema_id`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.

            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.

        Returns:
            ~.gp_schema.Schema:
                A schema resource.
        """
        # Create or coerce a protobuf request object.
        # Sanity check: If we got a request object, we should *not* have
        # gotten any keyword arguments that map to the request.
        has_flattened_params = any([parent, schema, schema_id])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        # Minor optimization to avoid making a copy if the user passes
        # in a gp_schema.CreateSchemaRequest.
        # There's no risk of modifying the input as we've already verified
        # there are no flattened fields.
        if not isinstance(request, gp_schema.CreateSchemaRequest):
            request = gp_schema.CreateSchemaRequest(request)

            # If we have keyword arguments corresponding to fields on the
            # request, apply these.

            if parent is not None:
                request.parent = parent
            if schema is not None:
                request.schema = schema
            if schema_id is not None:
                request.schema_id = schema_id

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = self._transport._wrapped_methods[self._transport.create_schema]

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("parent", request.parent),)),
        )

        # Send the request.
        response = rpc(request, retry=retry, timeout=timeout, metadata=metadata,)

        # Done; return the response.
        return response

    def get_schema(
        self,
        request: schema.GetSchemaRequest = None,
        *,
        name: str = None,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> schema.Schema:
        r"""Gets a schema.

        Args:
            request (:class:`~.schema.GetSchemaRequest`):
                The request object. Request for the GetSchema method.
            name (:class:`str`):
                Required. The name of the schema to get. Format is
                ``projects/{project}/schemas/{schema}``.
                This corresponds to the ``name`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.

            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.

        Returns:
            ~.schema.Schema:
                A schema resource.
        """
        # Create or coerce a protobuf request object.
        # Sanity check: If we got a request object, we should *not* have
        # gotten any keyword arguments that map to the request.
        has_flattened_params = any([name])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        # Minor optimization to avoid making a copy if the user passes
        # in a schema.GetSchemaRequest.
        # There's no risk of modifying the input as we've already verified
        # there are no flattened fields.
        if not isinstance(request, schema.GetSchemaRequest):
            request = schema.GetSchemaRequest(request)

            # If we have keyword arguments corresponding to fields on the
            # request, apply these.

            if name is not None:
                request.name = name

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = self._transport._wrapped_methods[self._transport.get_schema]

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", request.name),)),
        )

        # Send the request.
        response = rpc(request, retry=retry, timeout=timeout, metadata=metadata,)

        # Done; return the response.
        return response

    def list_schemas(
        self,
        request: schema.ListSchemasRequest = None,
        *,
        parent: str = None,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> pagers.ListSchemasPager:
        r"""Lists schemas in a project.

        Args:
            request (:class:`~.schema.ListSchemasRequest`):
                The request object. Request for the `ListSchemas`
                method.
            parent (:class:`str`):
                Required. The name of the project in which to list
                schemas. Format is ``projects/{project-id}``.
                This corresponds to the ``parent`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.

            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.

        Returns:
            ~.pagers.ListSchemasPager:
                Response for the ``ListSchemas`` method.

                Iterating over this object will yield results and
                resolve additional pages automatically.

        """
        # Create or coerce a protobuf request object.
        # Sanity check: If we got a request object, we should *not* have
        # gotten any keyword arguments that map to the request.
        has_flattened_params = any([parent])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        # Minor optimization to avoid making a copy if the user passes
        # in a schema.ListSchemasRequest.
        # There's no risk of modifying the input as we've already verified
        # there are no flattened fields.
        if not isinstance(request, schema.ListSchemasRequest):
            request = schema.ListSchemasRequest(request)

            # If we have keyword arguments corresponding to fields on the
            # request, apply these.

            if parent is not None:
                request.parent = parent

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = self._transport._wrapped_methods[self._transport.list_schemas]

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("parent", request.parent),)),
        )

        # Send the request.
        response = rpc(request, retry=retry, timeout=timeout, metadata=metadata,)

        # This method is paged; wrap the response in a pager, which provides
        # an `__iter__` convenience method.
        response = pagers.ListSchemasPager(
            method=rpc, request=request, response=response, metadata=metadata,
        )

        # Done; return the response.
        return response

    def delete_schema(
        self,
        request: schema.DeleteSchemaRequest = None,
        *,
        name: str = None,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> None:
        r"""Deletes a schema.

        Args:
            request (:class:`~.schema.DeleteSchemaRequest`):
                The request object. Request for the `DeleteSchema`
                method.
            name (:class:`str`):
                Required. Name of the schema to delete. Format is
                ``projects/{project}/schemas/{schema}``.
                This corresponds to the ``name`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.

            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        """
        # Create or coerce a protobuf request object.
        # Sanity check: If we got a request object, we should *not* have
        # gotten any keyword arguments that map to the request.
        has_flattened_params = any([name])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        # Minor optimization to avoid making a copy if the user passes
        # in a schema.DeleteSchemaRequest.
        # There's no risk of modifying the input as we've already verified
        # there are no flattened fields.
        if not isinstance(request, schema.DeleteSchemaRequest):
            request = schema.DeleteSchemaRequest(request)

            # If we have keyword arguments corresponding to fields on the
            # request, apply these.

            if name is not None:
                request.name = name

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = self._transport._wrapped_methods[self._transport.delete_schema]

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", request.name),)),
        )

        # Send the request.
        rpc(
            request, retry=retry, timeout=timeout, metadata=metadata,
        )

    def validate_schema(
        self,
        request: gp_schema.ValidateSchemaRequest = None,
        *,
        parent: str = None,
        schema: gp_schema.Schema = None,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> gp_schema.ValidateSchemaResponse:
        r"""Validates a schema.

        Args:
            request (:class:`~.gp_schema.ValidateSchemaRequest`):
                The request object. Request for the `ValidateSchema`
                method.
            parent (:class:`str`):
                Required. The name of the project in which to validate
                schemas. Format is ``projects/{project-id}``.
                This corresponds to the ``parent`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            schema (:class:`~.gp_schema.Schema`):
                Required. The schema object to
                validate.
                This corresponds to the ``schema`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.

            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.

        Returns:
            ~.gp_schema.ValidateSchemaResponse:
                Response for the ``ValidateSchema`` method.
        """
        # Create or coerce a protobuf request object.
        # Sanity check: If we got a request object, we should *not* have
        # gotten any keyword arguments that map to the request.
        has_flattened_params = any([parent, schema])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        # Minor optimization to avoid making a copy if the user passes
        # in a gp_schema.ValidateSchemaRequest.
        # There's no risk of modifying the input as we've already verified
        # there are no flattened fields.
        if not isinstance(request, gp_schema.ValidateSchemaRequest):
            request = gp_schema.ValidateSchemaRequest(request)

            # If we have keyword arguments corresponding to fields on the
            # request, apply these.

            if parent is not None:
                request.parent = parent
            if schema is not None:
                request.schema = schema

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = self._transport._wrapped_methods[self._transport.validate_schema]

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("parent", request.parent),)),
        )

        # Send the request.
        response = rpc(request, retry=retry, timeout=timeout, metadata=metadata,)

        # Done; return the response.
        return response

    def validate_message(
        self,
        request: schema.ValidateMessageRequest = None,
        *,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> schema.ValidateMessageResponse:
        r"""Validates a message against a schema.

        Args:
            request (:class:`~.schema.ValidateMessageRequest`):
                The request object. Request for the `ValidateMessage`
                method.

            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.

        Returns:
            ~.schema.ValidateMessageResponse:
                Response for the ``ValidateMessage`` method.
        """
        # Create or coerce a protobuf request object.

        # Minor optimization to avoid making a copy if the user passes
        # in a schema.ValidateMessageRequest.
        # There's no risk of modifying the input as we've already verified
        # there are no flattened fields.
        if not isinstance(request, schema.ValidateMessageRequest):
            request = schema.ValidateMessageRequest(request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = self._transport._wrapped_methods[self._transport.validate_message]

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("parent", request.parent),)),
        )

        # Send the request.
        response = rpc(request, retry=retry, timeout=timeout, metadata=metadata,)

        # Done; return the response.
        return response

    def set_iam_policy(
        self,
        request: iam_policy.SetIamPolicyRequest = None,
        *,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> policy.Policy:
        r"""Sets the IAM access control policy on the specified
        function. Replaces any existing policy.
        Args:
            request (:class:`~.iam_policy.SetIamPolicyRequest`):
                The request object. Request message for `SetIamPolicy`
                method.
            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.policy.Policy:
                Defines an Identity and Access Management (IAM) policy.
                It is used to specify access control policies for Cloud
                Platform resources.
                A ``Policy`` is a collection of ``bindings``. A
                ``binding`` binds one or more ``members`` to a single
                ``role``. Members can be user accounts, service
                accounts, Google groups, and domains (such as G Suite).
                A ``role`` is a named list of permissions (defined by
                IAM or configured by users). A ``binding`` can
                optionally specify a ``condition``, which is a logic
                expression that further constrains the role binding
                based on attributes about the request and/or target
                resource.
                **JSON Example**
                ::
                    {
                      "bindings": [
                        {
                          "role": "roles/resourcemanager.organizationAdmin",
                          "members": [
                            "user:mike@example.com",
                            "group:admins@example.com",
                            "domain:google.com",
                            "serviceAccount:my-project-id@appspot.gserviceaccount.com"
                          ]
                        },
                        {
                          "role": "roles/resourcemanager.organizationViewer",
                          "members": ["user:eve@example.com"],
                          "condition": {
                            "title": "expirable access",
                            "description": "Does not grant access after Sep 2020",
                            "expression": "request.time <
                            timestamp('2020-10-01T00:00:00.000Z')",
                          }
                        }
                      ]
                    }
                **YAML Example**
                ::
                    bindings:
                    - members:
                      - user:mike@example.com
                      - group:admins@example.com
                      - domain:google.com
                      - serviceAccount:my-project-id@appspot.gserviceaccount.com
                      role: roles/resourcemanager.organizationAdmin
                    - members:
                      - user:eve@example.com
                      role: roles/resourcemanager.organizationViewer
                      condition:
                        title: expirable access
                        description: Does not grant access after Sep 2020
                        expression: request.time < timestamp('2020-10-01T00:00:00.000Z')
                For a description of IAM and its features, see the `IAM
                developer's
                guide <https://cloud.google.com/iam/docs>`__.
        """
        # Create or coerce a protobuf request object.

        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = iam_policy.SetIamPolicyRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.set_iam_policy,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("resource", request.resource),)),
        )

        # Send the request.
        response = rpc(request, retry=retry, timeout=timeout, metadata=metadata,)

        # Done; return the response.
        return response

    def get_iam_policy(
        self,
        request: iam_policy.GetIamPolicyRequest = None,
        *,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> policy.Policy:
        r"""Gets the IAM access control policy for a function.
        Returns an empty policy if the function exists and does
        not have a policy set.
        Args:
            request (:class:`~.iam_policy.GetIamPolicyRequest`):
                The request object. Request message for `GetIamPolicy`
                method.
            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.policy.Policy:
                Defines an Identity and Access Management (IAM) policy.
                It is used to specify access control policies for Cloud
                Platform resources.
                A ``Policy`` is a collection of ``bindings``. A
                ``binding`` binds one or more ``members`` to a single
                ``role``. Members can be user accounts, service
                accounts, Google groups, and domains (such as G Suite).
                A ``role`` is a named list of permissions (defined by
                IAM or configured by users). A ``binding`` can
                optionally specify a ``condition``, which is a logic
                expression that further constrains the role binding
                based on attributes about the request and/or target
                resource.
                **JSON Example**
                ::
                    {
                      "bindings": [
                        {
                          "role": "roles/resourcemanager.organizationAdmin",
                          "members": [
                            "user:mike@example.com",
                            "group:admins@example.com",
                            "domain:google.com",
                            "serviceAccount:my-project-id@appspot.gserviceaccount.com"
                          ]
                        },
                        {
                          "role": "roles/resourcemanager.organizationViewer",
                          "members": ["user:eve@example.com"],
                          "condition": {
                            "title": "expirable access",
                            "description": "Does not grant access after Sep 2020",
                            "expression": "request.time <
                            timestamp('2020-10-01T00:00:00.000Z')",
                          }
                        }
                      ]
                    }
                **YAML Example**
                ::
                    bindings:
                    - members:
                      - user:mike@example.com
                      - group:admins@example.com
                      - domain:google.com
                      - serviceAccount:my-project-id@appspot.gserviceaccount.com
                      role: roles/resourcemanager.organizationAdmin
                    - members:
                      - user:eve@example.com
                      role: roles/resourcemanager.organizationViewer
                      condition:
                        title: expirable access
                        description: Does not grant access after Sep 2020
                        expression: request.time < timestamp('2020-10-01T00:00:00.000Z')
                For a description of IAM and its features, see the `IAM
                developer's
                guide <https://cloud.google.com/iam/docs>`__.
        """
        # Create or coerce a protobuf request object.

        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = iam_policy.GetIamPolicyRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.get_iam_policy,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("resource", request.resource),)),
        )

        # Send the request.
        response = rpc(request, retry=retry, timeout=timeout, metadata=metadata,)

        # Done; return the response.
        return response

    def test_iam_permissions(
        self,
        request: iam_policy.TestIamPermissionsRequest = None,
        *,
        retry: retries.Retry = gapic_v1.method.DEFAULT,
        timeout: float = None,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> iam_policy.TestIamPermissionsResponse:
        r"""Tests the specified permissions against the IAM access control
        policy for a function. If the function does not exist, this will
        return an empty set of permissions, not a NOT_FOUND error.
        Args:
            request (:class:`~.iam_policy.TestIamPermissionsRequest`):
                The request object. Request message for
                `TestIamPermissions` method.
            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.iam_policy.TestIamPermissionsResponse:
                Response message for ``TestIamPermissions`` method.
        """
        # Create or coerce a protobuf request object.

        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = iam_policy.TestIamPermissionsRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.test_iam_permissions,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("resource", request.resource),)),
        )

        # Send the request.
        response = rpc(request, retry=retry, timeout=timeout, metadata=metadata,)

        # Done; return the response.
        return response


try:
    DEFAULT_CLIENT_INFO = gapic_v1.client_info.ClientInfo(
        client_library_version=pkg_resources.get_distribution(
            "google-cloud-pubsub",
        ).version,
    )
except pkg_resources.DistributionNotFound:
    DEFAULT_CLIENT_INFO = gapic_v1.client_info.ClientInfo()


__all__ = ("SchemaServiceClient",)
